"""
Signals are a simple pub-sub mechanism.

DOMNodes can subscribe to a signal, which will invoke a callback when the signal is published.

This is experimental for now, for internal use. It may be part of the public API in a future release.

"""

from __future__ import annotations

from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

import rich.repr

from textual import log

if TYPE_CHECKING:
    from ._types import IgnoreReturnCallbackType
    from .dom import DOMNode


class SignalError(Exception):
    """Raised for Signal errors."""


@rich.repr.auto(angular=True)
class Signal:
    """A signal that a widget may subscribe to, in order to invoke callbacks when an associated event occurs."""

    def __init__(self, owner: DOMNode, name: str) -> None:
        """Initialize a signal.

        Args:
            owner: The owner of this signal.
            name: An identifier for debugging purposes.
        """
        self._owner = owner
        self._name = name
        self._subscriptions: WeakKeyDictionary[
            DOMNode, list[IgnoreReturnCallbackType]
        ] = WeakKeyDictionary()

    def __rich_repr__(self) -> rich.repr.Result:
        yield "owner", self._owner
        yield "name", self._name
        yield "subscriptions", list(self._subscriptions.keys())

    def subscribe(self, node: DOMNode, callback: IgnoreReturnCallbackType) -> None:
        """Subscribe a node to this signal.

        When the signal is published, the callback will be invoked.

        Args:
            node: Node to subscribe.
            callback: A callback function which takes no arguments, and returns anything (return type ignored).

        Raises:
            SignalError: Raised when subscribing a non-mounted widget.
        """
        if not node.is_running:
            raise SignalError(
                f"Node must be running to subscribe to a signal (has {node} been mounted)?"
            )
        callbacks = self._subscriptions.setdefault(node, [])
        if callback not in callbacks:
            callbacks.append(callback)

    def unsubscribe(self, node: DOMNode) -> None:
        """Unsubscribe a node from this signal.

        Args:
            node: Node to unsubscribe,
        """
        self._subscriptions.pop(node, None)

    def publish(self) -> None:
        """Publish the signal (invoke subscribed callbacks)."""

        for node, callbacks in list(self._subscriptions.items()):
            if not node.is_running:
                # Removed nodes that are no longer running
                self._subscriptions.pop(node)
            else:
                # Call callbacks
                for callback in callbacks:
                    try:
                        callback()
                    except Exception as error:
                        log.error(f"error publishing signal to {node} ignored; {error}")



import pytest

from textual.app import App, ComposeResult
from textual.signal import Signal, SignalError
from textual.widgets import Label


async def test_signal():
    """Test signal subscribe"""
    called = 0

    class TestLabel(Label):
        def on_mount(self) -> None:
            def signal_result():
                nonlocal called
                called += 1

            assert isinstance(self.app, TestApp)
            self.app.test_signal.subscribe(self, signal_result)

    class TestApp(App):
        BINDINGS = [("space", "signal")]

        def __init__(self) -> None:
            self.test_signal = Signal(self, "coffee ready")
            super().__init__()

        def compose(self) -> ComposeResult:
            yield TestLabel()

        def action_signal(self) -> None:
            self.test_signal.publish()

    app = TestApp()
    async with app.run_test() as pilot:
        # Check default called is 0
        assert called == 0
        # Action should publish signal
        await pilot.press("space")
        assert called == 1
        # Check a second time
        await pilot.press("space")
        assert called == 2
        # Removed the owner object
        await app.query_one(TestLabel).remove()
        # Check nothing is called
        await pilot.press("space")
        assert called == 2
        # Add a new test label
        await app.mount(TestLabel())
        # Check callback again
        await pilot.press("space")
        assert called == 3
        # Unsubscribe
        app.test_signal.unsubscribe(app.query_one(TestLabel))
        # Check nothing to update
        await pilot.press("space")
        assert called == 3


def test_signal_errors():
    """Check exceptions raised by Signal class."""
    app = App()
    test_signal = Signal(app, "test")
    label = Label()
    # Check subscribing a non-running widget is an error
    with pytest.raises(SignalError):
        test_signal.subscribe(label, lambda: None)


def test_repr():
    """Check the repr doesn't break."""
    app = App()
    test_signal = Signal(app, "test")
    assert isinstance(repr(test_signal), str)
